/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___PG_ConfigPanel___
#define ___PG_ConfigPanel___

namespace ProceduralTextures
{
	class SpecificsBase
	{
	public:
		eCONTROL_TYPE	m_eCtrlType;

	public:
		SpecificsBase( eCONTROL_TYPE p_eCtrlType )
			:	m_eCtrlType	( p_eCtrlType )
		{
		}
		virtual ~SpecificsBase()
		{
		}
	};
	/*!
	\author Sylvain DOREMUS
	\date 23/05/2012
	\~english
	\brief	Specialisation of Specifics for wxStaticText
	\~french
	\brief	Spécialisation de Specifics pour les wxStaticText
	*/
	template<> class Specifics< eCONTROL_TYPE_STATIC > : public SpecificsBase
	{
	public:
		wxString		m_strText;
		wxStaticText *	m_pControl;

	public:
		Specifics( wxString const & p_strText )
			:	SpecificsBase( eCONTROL_TYPE_STATIC )
			,	m_strText	( p_strText	)
			,	m_pControl	( NULL		)
		{
		}
	};
	/*!
	\author Sylvain DOREMUS
	\date 23/05/2012
	\~english
	\brief	Specialisation of Specifics for wxTextCtrl
	\~french
	\brief	Spécialisation de Specifics pour les wxTextCtrl
	*/
	template<> class Specifics< eCONTROL_TYPE_EDIT > : public SpecificsBase
	{
	public:
		wxString		m_strText;
		wxTextCtrl *	m_pControl;

	public:
		Specifics( wxString const & p_strText )
			:	SpecificsBase( eCONTROL_TYPE_EDIT )
			,	m_strText	( p_strText	)
			,	m_pControl	( NULL		)
		{
		}
	};
	/*!
	\author Sylvain DOREMUS
	\date 23/05/2012
	\~english
	\brief	Specialisation of Specifics for wxSlider
	\~french
	\brief	Spécialisation de Specifics pour les wxSlider
	*/
	template<> class Specifics< eCONTROL_TYPE_SLIDER > : public SpecificsBase
	{
	public:
		int			m_iMin;
		int			m_iMax;
		int			m_iValue;
		wxSlider *	m_pControl;

	public:
		Specifics( int p_iMin, int p_iMax, int p_iValue )
			:	SpecificsBase( eCONTROL_TYPE_SLIDER )
			,	m_iMin		( p_iMin	)
			,	m_iMax		( p_iMax	)
			,	m_iValue	( p_iValue	)
			,	m_pControl	( NULL		)
		{
		}
	};
	/*!
	\author Sylvain DOREMUS
	\date 23/05/2012
	\~english
	\brief	Specialisation of Specifics for wxComboBox
	\~french
	\brief	Spécialisation de Specifics pour les wxComboBox
	*/
	template<> class Specifics< eCONTROL_TYPE_COMBO > : public SpecificsBase
	{
	public:
		wxString		m_strValue;
		wxArrayString	m_arrayValues;
		wxComboBox *	m_pControl;

	public:
		Specifics( wxString const & p_strValue, wxArrayString const & p_arrayValues )
			:	SpecificsBase( eCONTROL_TYPE_COMBO )
			,	m_strValue		( p_strValue	)
			,	m_arrayValues	( p_arrayValues	)
			,	m_pControl		( NULL			)
		{
		}
		template< size_t N >
		Specifics( wxString const & p_strValue, wxString const (& p_arrayValues)[N] )
			:	SpecificsBase( eCONTROL_TYPE_COMBO )
			,	m_strValue		( p_strValue	)
			,	m_pControl		( NULL			)
		{
			m_arrayValues.insert( m_arrayValues.end(), &p_arrayValues[0], &p_arrayValues[N] );
		}
		Specifics( wxString const & p_strValue, wxString const * p_arrayValues, size_t p_uiCount )
			:	SpecificsBase( eCONTROL_TYPE_COMBO )
			,	m_strValue		( p_strValue	)
			,	m_pControl		( NULL			)
		{
			m_arrayValues.insert( m_arrayValues.end(), &p_arrayValues[0], &p_arrayValues[p_uiCount] );
		}
	};
	/*!
	\author Sylvain DOREMUS
	\date 23/05/2012
	\~english
	\brief	Specialisation of Specifics for wxButton
	\~french
	\brief	Spécialisation de Specifics pour les wxButton
	*/
	template<> class Specifics< eCONTROL_TYPE_BUTTON > : public SpecificsBase
	{
	public:
		wxString	m_strText;
		wxButton *	m_pControl;

	public:
		Specifics( wxString const & p_strText )
			:	SpecificsBase( eCONTROL_TYPE_BUTTON )
			,	m_strText	( p_strText	)
			,	m_pControl	( NULL		)
		{
		}
	};

	class ControlInfos
	{
	public:

	public:
		eCONTROL_TYPE			m_eType;		//!<\~english The control type												\~french Le type de contrôle
		wxWindowID				m_iId;			//!<\~english The control ID												\~french L'ID du contrôle
		wxPoint					m_ptPosition;	//!<\~english The control position											\~french La position du contrôle
		wxSize					m_szSize;		//!<\~english The control dimensions										\~french Les dimensions du contrôle
		SpecificsBase *			m_pSpecific;	//!<\~english Control's specific data										\~french Des données spécifiques au contrôle
		int						m_iStyle;		//!<\~english The control style												\~french Le style du contrôle
		eEVENT_TYPE				m_eEventType;	//!<\~english An event on which a function is called						\~french Un évènement sur lequel le contrôle doit réagir
		wxObjectEventFunction	m_pfnFunction;	//!<\~english Function to use for specific event launched by the control	\~french La fonction de réaction sur l'évènement
		bool					m_bVisible;		//!<\~english Initial visibility status										\~french Statut de visibilité initiale

	public:
		/**
		 *\~english
		 *\brief		Default constructor
		 *\param[in]	p_eType			The control type
		 *\param[in]	p_iId			The control ID
		 *\param[in]	p_ptPosition	The control position
		 *\param[in]	p_szSize		The control dimensions
		 *\param[in]	p_pSpecific		Control's specific data
		 *\param[in]	p_iStyle		The control style
		 *\param[in]	p_event			An event on which a function is called
		 *\param[in]	p_pfnFunction	Function to use for specific event launched by the control
		 *\param[in]	p_bVisible		Initial visibility status
		 *\~french
		 *\brief		Constructeur par défaut
		 *\param[in]	p_eType			Le type de contrôle
		 *\param[in]	p_iId			L'ID du contrôle
		 *\param[in]	p_ptPosition	La position du contrôle
		 *\param[in]	p_szSize		Les dimensions du contrôle
		 *\param[in]	p_pSpecific		Des données spécifiques au contrôle
		 *\param[in]	p_iStyle		Le style du contrôle
		 *\param[in]	p_event			Un évènement sur lequel le contrôle doit réagir
		 *\param[in]	p_pfnFunction	La fonction de réaction sur l'évènement
		 *\param[in]	p_bVisible		Statut de visibilité initiale
		 */
		ControlInfos( eCONTROL_TYPE p_eType, wxWindowID p_iId, wxPoint const & p_ptPosition, wxSize const & p_szSize, SpecificsBase * p_pSpecific, int p_iStyle=0, eEVENT_TYPE p_eEventType=eEVENT_TYPE_NONE, wxObjectEventFunction p_pfnFunction=NULL, bool p_bVisible=true );
		/**
		 *\~english
		 *\brief		Copy constructor
		 *\param[in]	p_copy	Element to copy
		 *\~french
		 *\brief		Constructeur par copie
		 *\param[in]	p_copy	Elément à copier
		 */
		ControlInfos( ControlInfos const & p_copy );
		/**
		 *\~english
		 *\brief		Move constructor
		 *\param[in]	p_copy	Element to move
		 *\~french
		 *\brief		Constructeur par déplacement
		 *\param[in]	p_copy	Elément à déplacer
		 */
		ControlInfos( ControlInfos && p_copy );
		/**
		 *\~english
		 *\brief		Copy assignment operator
		 *\param[in]	p_copy	Element to copy
		 *\return		A reference over the current element
		 *\~french
		 *\brief		Opérateur d'affectation par copie
		 *\param[in]	p_copy	Elément à copier
		 *\return		Une référence sur l'élément courant
		 */
		ControlInfos & operator =( ControlInfos const & p_copy );
		/**
		 *\~english
		 *\brief		Move assignment operator
		 *\param[in]	p_copy	Element to move
		 *\return		A reference over the current element
		 *\~french
		 *\brief		Opérateur d'affectation par déplacement
		 *\param[in]	p_copy	Elément à déplacer
		 *\return		Une référence sur l'élément courant
		 */
		ControlInfos & operator =( ControlInfos && p_copy );
		/**
		 *\~english
		 *\brief		Destructor
		 *\~french
		 *\brief		Destructeur
		 */
		virtual ~ControlInfos();
	};
	/*!
	\author Sylvain DOREMUS
	\date 23/05/2012
	\~english
	\brief	Typedef on an array of ControlInfos
	\~french
	\brief	Typedef sur un tableau de ControlInfos
	*/
	typedef std::vector< ControlInfos > CtrlInfosArray;

	class ConfigPanel
        :   public wxPanel
	{
	private:
		ProceduralGenerator *	m_pGenerator;						//!<\~english The generator									\~french Le générateur
		wxEventType				m_arrayTypes[eEVENT_TYPE_COUNT];	//!<\~english Array connecting wxEventType and eEVENT_TYPE 	\~french Tableau de correspondances entre eEVENT_TYPE et wxEventType

	public:
		ConfigPanel( wxEventType const (&p_arrayTypes)[eEVENT_TYPE_COUNT], ProceduralGenerator * p_pGenerator, wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition=wxDefaultPosition, const wxSize & p_size=wxDefaultSize);
	    virtual ~ConfigPanel();

		void CreateSpecifics();

	private:
		void DoCreateStatic		( ControlInfos & p_ctrlInfos );
		void DoCreateEdit		( ControlInfos & p_ctrlInfos );
		void DoCreateSlider		( ControlInfos & p_ctrlInfos );
		void DoCreateCombo		( ControlInfos & p_ctrlInfos );
		void DoCreateButton		( ControlInfos & p_ctrlInfos );

		void OnLoadImageA		( wxCommandEvent	&	p_event );
		void OnLoadImageB		( wxCommandEvent	&	p_event );
		void OnLoadTextFileA	( wxCommandEvent	&	p_event );
		void OnLoadTextFileB	( wxCommandEvent	&	p_event );
		void OnLoadBinaryFileA	( wxCommandEvent	&	p_event );
		void OnLoadBinaryFileB	( wxCommandEvent	&	p_event );
		void OnLoadColourA		( wxCommandEvent	&	p_event );
		void OnLoadColourB		( wxCommandEvent	&	p_event );
	};
}

#endif
