/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___Pg_GlObjects___
#define ___Pg_GlObjects___

#include "glew.h"

#include "Vertex.h"
#include "Pixel.h"

#include <list>

//*************************************************************************************************

extern bool CheckGlError( const wxString & p_strMsg);

typedef void	(wxCALLBACK * PFNGLCLEARCOLORPROC)			( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha);
typedef void	(wxCALLBACK * PFNGLSHADEMODELPROC)			( GLenum mode);
typedef void	(wxCALLBACK * PFNGLVIEWPORTPROC)			( GLint x, GLint y, GLsizei width, GLsizei height);
typedef void	(wxCALLBACK * PFNGLMATRIXMODEPROC)			( GLenum mode);
typedef void	(wxCALLBACK * PFNGLLOADIDENTITYPROC)		( void);
typedef void	(wxCALLBACK * PFNGLORTHOPROC)				( GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar);
typedef void	(wxCALLBACK * PFNGLCLEARPROC)				( GLbitfield mask);
typedef void	(wxCALLBACK * PFNGLDRAWELEMENTSPROC)		( GLenum mode, GLsizei count, GLenum type, const GLvoid *indices);
typedef void	(wxCALLBACK * PFNGLENABLEPROC)				( GLenum mode);
typedef void	(wxCALLBACK * PFNGLDISABLEPROC)				( GLenum mode);
typedef void	(wxCALLBACK * PFNGLGENTEXTURESPROC)			( GLsizei n, GLuint * textures);
typedef void	(wxCALLBACK * PFNGLDELETETEXTURESPROC)		( GLsizei n, const GLuint * textures);
typedef void	(wxCALLBACK * PFNGLBINDTEXTUREPROC)			( GLenum mode, GLuint texture);
typedef void	(wxCALLBACK * PFNGLTEXSUBIMAGE2DPROC)		( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid * data);
typedef void	(wxCALLBACK * PFNGLTEXIMAGE2DPROC)			( GLenum target, GLint level, GLint internalFormat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid * data);
typedef void	(wxCALLBACK * PFNGLTEXPARAMETERIPROC)		( GLenum target, GLenum pname, GLint param);
typedef void	(wxCALLBACK * PFNGLGETTEXIMAGEPROC)			( GLenum target, GLint level, GLenum format, GLenum type, GLvoid * pixels);
typedef void	(wxCALLBACK * PFNGLENABLECLIENTSTATEPROC)	( GLenum array);
typedef void	(wxCALLBACK * PFNGLDISABLECLIENTSTATEPROC)	( GLenum array);
typedef void	(wxCALLBACK * PFNGLVERTEXPOINTERPROC)		( GLint size, GLenum type, GLsizei stride, const GLvoid * pointer);
typedef void	(wxCALLBACK * PFNGLNORMALPOINTERPROC)		( GLenum type, GLsizei stride, const GLvoid * pointer);
typedef void	(wxCALLBACK * PFNGLTEXCOORDPOINTERPROC)		( GLint size, GLenum type, GLsizei stride, const GLvoid  *pointer);
typedef void	(wxCALLBACK * PFNGLTEXENVIPROC)				( GLenum target, GLenum pname, GLint param);
typedef void	(wxCALLBACK * PFNGLTEXENVIVPROC)			( GLenum target, GLenum pname, const GLint * param);
typedef void	(wxCALLBACK * PFNGLTEXENVFPROC)				( GLenum target, GLenum pname, GLfloat param);
typedef void	(wxCALLBACK * PFNGLTEXENVFVPROC)			( GLenum target, GLenum pname, const GLfloat * param);
typedef void	(wxCALLBACK * PFNGLBLENDFUNCPROC)			( GLenum sfactor, GLenum dfactor);
typedef void	(wxCALLBACK * PFNGLALPHAFUNCPROC)			( GLenum func, GLclampf ref);
typedef GLenum	(wxCALLBACK * PFNGLGETERRORPROC)			( void);
typedef void	(wxCALLBACK * PFNGLREADBUFFERPROC)			( GLenum mode);
typedef void	(wxCALLBACK * PFNGLREADPIXELSPROC)			( GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels);

//*************************************************************************************************

namespace ProceduralTextures
{
	class OpenGl
	{
	private:
		/**@name General */
		//@{
		PFNGLGETERRORPROC					m_pfnGetError;
		PFNGLCLEARCOLORPROC					m_pfnClearColor;
		PFNGLSHADEMODELPROC					m_pfnShadeModel;
		PFNGLVIEWPORTPROC					m_pfnViewport;
		PFNGLMATRIXMODEPROC					m_pfnMatrixMode;
		PFNGLLOADIDENTITYPROC				m_pfnLoadIdentity;
		PFNGLORTHOPROC						m_pfnOrtho;
		PFNGLCLEARPROC						m_pfnClear;
		PFNGLDRAWELEMENTSPROC				m_pfnDrawElements;
		PFNGLENABLEPROC						m_pfnEnable;
		PFNGLDISABLEPROC					m_pfnDisable;
		//@}

		/**@name Texture */
		//@{
		PFNGLGENTEXTURESPROC				m_pfnGenTextures;
		PFNGLDELETETEXTURESPROC				m_pfnDeleteTextures;
		PFNGLBINDTEXTUREPROC				m_pfnBindTexture;
		PFNGLTEXSUBIMAGE2DPROC				m_pfnTexSubImage2D;
		PFNGLTEXIMAGE2DPROC					m_pfnTexImage2D;
		PFNGLTEXPARAMETERIPROC				m_pfnTexParameteri;
		PFNGLGENERATEMIPMAPPROC				m_pfnGenerateMipmap;
		PFNGLACTIVETEXTUREPROC				m_pfnActiveTexture;
		PFNGLCLIENTACTIVETEXTUREPROC		m_pfnClientActiveTexture;
		PFNGLGETTEXIMAGEPROC				m_pfnGetTexImage;
		PFNGLTEXENVIPROC					m_pfnTexEnvi;
		PFNGLTEXENVIVPROC					m_pfnTexEnviv;
		PFNGLTEXENVFPROC					m_pfnTexEnvf;
		PFNGLTEXENVFVPROC					m_pfnTexEnvfv;
		PFNGLBLENDFUNCPROC					m_pfnBlendFunc;
		PFNGLALPHAFUNCPROC					m_pfnAlphaFunc;
		PFNGLREADBUFFERPROC					m_pfnReadBuffer;
		PFNGLREADPIXELSPROC					m_pfnReadPixels;
		//@}

		/**@name Buffer Objects */
		//@{
		PFNGLENABLECLIENTSTATEPROC			m_pfnEnableClientState;
		PFNGLDISABLECLIENTSTATEPROC			m_pfnDisableClientState;
		PFNGLGENBUFFERSPROC					m_pfnGenBuffers;
		PFNGLBINDBUFFERPROC					m_pfnBindBuffer;
		PFNGLDELETEBUFFERSPROC				m_pfnDeleteBuffers;
		PFNGLISBUFFERPROC					m_pfnIsBuffer;
		PFNGLBUFFERDATAPROC					m_pfnBufferData;
		PFNGLBUFFERSUBDATAARBPROC			m_pfnBufferSubData;
		PFNGLGETBUFFERPARAMETERIVPROC		m_pfnGetBufferParameteriv;
		PFNGLMAPBUFFERPROC					m_pfnMapBuffer;
		PFNGLUNMAPBUFFERPROC				m_pfnUnmapBuffer;
		PFNGLVERTEXPOINTERPROC				m_pfnVertexPointer;
		PFNGLNORMALPOINTERPROC				m_pfnNormalPointer;
		PFNGLTEXCOORDPOINTERPROC			m_pfnTexCoordPointer;
		//@}

		/**@name FBO */
		//@{
		PFNGLGENFRAMEBUFFERSPROC			m_pfnGenFramebuffers;
		PFNGLDELETEFRAMEBUFFERSPROC			m_pfnDeleteFramebuffers;
		PFNGLBINDFRAMEBUFFERPROC			m_pfnBindFramebuffer;
		PFNGLFRAMEBUFFERTEXTURE2DPROC		m_pfnFramebufferTexture2D;
		PFNGLCHECKFRAMEBUFFERSTATUSPROC		m_pfnCheckFramebufferStatus;
		//@}

		/**@name Uniform variables */
		//@{
		PFNGLGETUNIFORMLOCATIONPROC			m_pfnGetUniformLocation;
		PFNGLUNIFORM1IPROC					m_pfnUniform1i;
		PFNGLUNIFORM1FPROC					m_pfnUniform1f;
		//@}

		/**@name Shader object */
		//@{
		PFNGLCREATESHADERPROC				m_pfnCreateShader;
		PFNGLDELETESHADERPROC				m_pfnDeleteShader;
		PFNGLATTACHSHADERPROC				m_pfnAttachShader;
		PFNGLDETACHSHADERPROC				m_pfnDetachShader;
		PFNGLCOMPILESHADERPROC				m_pfnCompileShader;
		PFNGLGETSHADERIVPROC				m_pfnGetShaderiv;
		PFNGLGETSHADERINFOLOGPROC			m_pfnGetShaderInfoLog;
		PFNGLSHADERSOURCEPROC				m_pfnShaderSource;
		//@}

		/**@name Shader program */
		//@{
		PFNGLCREATEPROGRAMPROC				m_pfnCreateProgram;
		PFNGLDELETEPROGRAMPROC				m_pfnDeleteProgram;
		PFNGLLINKPROGRAMPROC				m_pfnLinkProgram;
		PFNGLUSEPROGRAMPROC					m_pfnUseProgram;
		PFNGLGETPROGRAMIVPROC				m_pfnGetProgramiv;
		PFNGLGETPROGRAMINFOLOGPROC			m_pfnGetProgramInfoLog;
		PFNGLGETATTRIBLOCATIONPROC			m_pfnGetAttribLocation;
		PFNGLISPROGRAMPROC					m_pfnIsProgram;
		//@}

		//! Private static instance, initialisable only one time
		static OpenGl * m_instance;

		/**@name Construction / Destruction => Private so only available through Initialise and Cleanup */
		//@{
		OpenGl();
		~OpenGl();
		//@}

	public:
		static void		Initialise	();
		static void		Cleanup		();

		/**@name General Functions */
		//@{
		static GLenum	GetError				();
		static bool		ClearColor				( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha);
		static bool		ShadeModel				( GLenum mode);
		static bool		Viewport				( GLint x, GLint y, GLsizei width, GLsizei height);
		static bool		MatrixMode				( GLenum mode);
		static bool		LoadIdentity			();
		static bool		Ortho					( GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar);
		static bool		Clear					( GLbitfield mask);
		static bool		DrawElements			( GLenum mode, GLsizei count, GLenum type, const GLvoid *indices);
		static bool		Enable					( GLenum mode);
		static bool		Disable					( GLenum mode);
		//@}

		/**@name Texture functions */
		//@{
		static bool		GenTextures				( GLsizei n, GLuint * textures);
		static bool		DeleteTextures			( GLsizei n, const GLuint * textures);
		static bool		GenerateMipmap			( GLenum target);
		static bool		ActiveTexture			( GLint texture);
		static bool		ClientActiveTexture		( GLenum texture);
		static bool		BindTexture				( GLenum mode, GLuint texture);
		static bool		TexSubImage2D			( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid * data);
		static bool		TexImage2D				( GLenum target, GLint level, GLint internalFormat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid * data);
		static bool		TexParameteri			( GLenum target, GLenum pname, GLint param);
		static bool		GetTexImage				( GLenum target, GLint level, GLenum format, GLenum type, GLvoid * img);
		static bool		TexEnvi					( GLenum target, GLenum pname, GLint param);
		static bool		TexEnviv				( GLenum target, GLenum pname, const GLint * param);
		static bool		TexEnvf					( GLenum target, GLenum pname, GLfloat param);
		static bool		TexEnvfv				( GLenum target, GLenum pname, const GLfloat * param);
		static bool		BlendFunc				( GLenum sfactor, GLenum dfactor);
		static bool		AlphaFunc				( GLenum func, GLclampf ref);
		static bool		ReadBuffer				( GLenum mode);
		static bool		ReadPixels				( GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid * pixels);
		//@}

		/**@name Buffer objects functions */
		//@{
		static bool		EnableClientState		( GLenum array);
		static bool		DisableClientState		( GLenum array);
		static bool		GenBuffers				( GLsizei n, GLuint* buffers);
		static bool		BindBuffer				( GLenum target, GLuint buffer);
		static bool		DeleteBuffers			( GLsizei n, const GLuint* buffers);
		static bool		IsBuffer				( GLuint buffer);
		static bool		BufferData				( GLenum target, GLsizeiptr size, const GLvoid* data, GLenum usage);
		static bool		BufferSubData			( GLenum target, GLintptrARB offset, GLsizeiptrARB size, const GLvoid * data);
		static GLvoid *	MapBuffer				( GLenum target, GLenum access);
		static bool		UnmapBuffer				( GLenum target);
		static bool		GetBufferParameteriv	( GLenum target, GLenum pname, GLint * params);
		static bool		VertexPointer			( GLint size, GLenum type, GLsizei stride, const GLvoid * pointer);
		static bool		NormalPointer			( GLenum type, GLsizei stride, const GLvoid * pointer);
		static bool		TexCoordPointer			( GLint size, GLenum type, GLsizei stride, const GLvoid * pointer);
		//@}

		/**@name FBO functions */
		//@{
		static bool		GenFramebuffers			( GLsizei n, GLuint* framebuffers);
		static bool		DeleteFramebuffers		( GLsizei n, const GLuint* framebuffers);
		static bool		BindFramebuffer			( GLenum target, GLuint framebuffer);
		static bool		FramebufferTexture2D	( GLenum target, GLenum attachment, GLenum textarget, GLuint texture, GLint level);
		static GLenum	CheckFramebufferStatus	( GLenum target);
		//@}

		/**@name Uniform variable Functions */
		//@{
		static GLint	GetUniformLocation		( GLuint program, const GLchar* name);
		static bool		Uniform1i				( GLint location, GLint v0);
		static bool		Uniform1f				( GLint location, GLfloat v0);
		//@}

		/**@name Shader object Functions */
		//@{
		static GLuint	CreateShader			( GLenum type);
		static bool		DeleteShader			( GLuint shader);
		static bool		AttachShader			( GLuint program, GLuint shader);
		static bool		DetachShader			( GLuint program, GLuint shader);
		static bool		CompileShader			( GLuint shader);
		static bool		GetShaderiv				( GLuint shader, GLenum pname, GLint* param);
		static bool		GetShaderInfoLog		( GLuint shader, GLsizei bufSize, GLsizei* length, GLchar* infoLog);
		static bool		ShaderSource			( GLuint shader, GLsizei count, const GLchar** strings, const GLint* lengths);
		//@}

		/**@name Shader program Functions */
		//@{
		static GLuint	CreateProgram			();
		static bool		DeleteProgram			( GLuint program);
		static bool		LinkProgram				( GLuint program);
		static bool		UseProgram				( GLuint program);
		static bool		GetProgramiv			( GLuint program, GLenum pname, GLint* param);
		static bool		GetProgramInfoLog		( GLuint program, GLsizei bufSize, GLsizei* length, GLchar* infoLog);
		static GLint	GetAttribLocation		( GLuint program, const GLchar* name);
		static bool		IsProgram				( GLuint program);
		//@}
	};

	class GlBuffer
	{
	private:
		GLuint	m_uiGlIndex;
		GLenum	m_target;
		GLenum	m_mode;

	public:
		GlBuffer( GLenum p_target, GLenum p_mode);
		virtual ~GlBuffer();

		void Create();
		void Cleanup();

		bool Bind();
		bool Unbind();
		void Data( void * p_pBuffer, size_t p_uiSize);
		void * Lock( GLenum p_access);
		bool Unlock();

		virtual void Initialise()=0;
		virtual bool Activate()=0;
		virtual void Deactivate()=0;

		inline GLuint GetGlIndex()const	{ return m_uiGlIndex; }
	};

	class GlVertexBuffer : public GlBuffer
	{
	private:
		BufferElementDeclaration *	m_vertexDeclarationElements;
		BufferDeclaration *			m_vertexDeclaration;
		std::vector< Vertex >		m_arrayVertex;
		GLfloat						m_pBuffer[44];//4 * [3(vertex position) + 3(vertex normal) + 3(vertex tangent) + 2(texture coordinates)]

	public:
		GlVertexBuffer();
		virtual ~GlVertexBuffer();

		virtual void Initialise();
		virtual bool Activate();
		virtual void Deactivate();
	};

	class GlIndexBuffer : public GlBuffer
	{
	private:
		static const int Size = 6;
		GLuint		m_pBuffer[Size];

	public:
		GlIndexBuffer();
		virtual ~GlIndexBuffer();

		virtual void Initialise();
		virtual bool Activate();
		virtual void Deactivate();

		inline int GetSize()const { return Size; }
	};

	class GlPixelBuffer : public GlBuffer
	{
	protected:
		PixelBuffer	* m_pPixels;
		GLenum m_packMode;
		GLenum m_drawMode;

	public:
		GlPixelBuffer( PixelBuffer * p_pPixels, GLenum p_packMode, GLenum p_drawMode);
		virtual ~GlPixelBuffer();

		virtual bool Activate();
		virtual void Deactivate();
	};

	class GlPackPixelBuffer : public GlPixelBuffer
	{
	public:
		GlPackPixelBuffer( PixelBuffer * p_pPixels);
		virtual ~GlPackPixelBuffer();

		virtual void Initialise();
	};

	class GlUnpackPixelBuffer : public GlPixelBuffer
	{
	public:
		GlUnpackPixelBuffer( PixelBuffer * p_pPixels);
		virtual ~GlUnpackPixelBuffer();

		virtual void Initialise();
	};

	class GlTexture
	{
	public:
		struct BlendMode
		{
			GLenum function;
			GLenum source1;
			GLenum source2;
			GLenum source3;
			GLenum operand1;
			GLenum operand2;
			GLenum operand3;
		};

	private:
		GLuint			m_uiGlTexture;
		bool			m_bInitialised;
		wxSize			m_size;
		PixelBuffer *	m_pPixels;
		bool			m_bOwnsPixels;
		GlPixelBuffer *	m_pIoBuffers[2];
		GlPixelBuffer *	m_pIoBuffer;
		Pixel<float>	m_pxColour;
		int				m_iCurrentPbo;

	public:
		GlTexture();
		~GlTexture();

		void Initialise();
		void Initialise( PixelBuffer * p_buffer);
		void Initialise( const wxSize & p_size);
		void Cleanup();

		bool Activate( GLenum p_texTarget=GL_TEXTURE0, GLenum p_mode=GL_MODULATE, BlendMode * p_blendRgb=NULL, BlendMode * p_blendAlpha=NULL, bool p_bUpload=false);
		void Deactivate( GLenum p_texTarget=GL_TEXTURE0);

		void UploadGlImage();
		void UploadGlImageSync();
		void DownloadGlImage();
/*
		void * LockGlImage( GLenum p_mode);
		void UnlockGlImage( GLenum p_mode);
*/
		void Resize( const wxSize & p_size);

		inline GLuint				GetGlIndex	()const	{ return m_uiGlTexture; }
		inline const PixelBuffer &	GetBuffer	()const	{ return ( * m_pPixels); }
		inline PixelBuffer &		GetBuffer	()		{ return ( * m_pPixels); }

		inline void SetColour	( const UbPixel & p_pxColour)		{ m_pxColour.Set( p_pxColour.r / 255.0f, p_pxColour.g / 255.0f, p_pxColour.b / 255.0f, p_pxColour.a / 255.0f); }
		inline void SetColour	( const Pixel<float> & p_pxColour)	{ m_pxColour = p_pxColour; }

	private:
		void DoInitGl();
		void DoDeleteBuffers();
		void DoCleanupPbos();
		void DoInitialisePbos();
	};

	class GlFrameBuffer
	{
	private:
		GLuint								m_uiGlBuffer;
		bool								m_bInitialised;
		ProceduralGenerator				*	m_pGenerator;
		wxSize								m_size;
		std::map<GLenum, GlTexture *>		m_mapAttachments;
		GlPixelBuffer					*	m_pPixelBuffer;

	public:
		GlFrameBuffer();
		virtual ~GlFrameBuffer();

		void Initialise();
		void Initialise( ProceduralGenerator * p_pGenerator);
		void Cleanup();
		bool Activate();
		void Deactivate();
		bool ActivateTexture( GLenum p_attachment);
		void DeactivateTexture( GLenum p_attachment);
		void BindTexture( GLenum p_attachment, GlTexture * p_pTexture);
		GlTexture * GetTexture( GLenum p_attachment);
/*
		void UploadTexture( GLenum p_attachment, const PixelBuffer & p_buffer);
		void DownloadTexture( GLenum p_attachment, PixelBuffer & p_buffer);
*/
		bool DownloadBuffer( GLenum p_attachment, PixelBuffer & p_buffer);

		void Resize( const wxSize & p_size);
	};

	class GlFrameVariableBase
	{
	protected:
		friend class GlShaderProgram;
		int						m_glIndex;
		GLuint				*	m_uiParentProgram;
		bool					m_bPresentInProgram;
		GlShaderProgram		*	m_pProgram;
		wxString				m_strName;
		bool					m_bChanged;

	protected:
		GlFrameVariableBase( GlShaderProgram * p_pProgram, GLuint * p_uiProgram);
		virtual ~GlFrameVariableBase();

	public:
		/**
		 * Assigns and activate the frame variable
		 */
		virtual bool Apply()=0;
		/**
		 * Retrieves the variable location in the shader program
		 */
		void GetVariableLocation( const GLchar * p_pVarName);

		inline wxString	GetName		()const	{ return m_strName; }
		inline int		GetGlIndex	()const	{ return m_glIndex; }

		inline void SetName		( const wxString & p_strName)	{ m_strName = p_strName; }
		inline void SetChanged	( bool p_bVal=true)				{ m_bChanged = p_bVal; }
	};
	template <typename T> struct VariableApplyer;
	template <typename T>
	class GlFrameVariable : public GlFrameVariableBase
	{
	protected:
		friend class GlShaderProgram;
		T	m_tValue;

	protected:
		GlFrameVariable( GlShaderProgram * p_pProgram, GLuint * p_uiProgram)
			:	GlFrameVariableBase( p_pProgram, p_uiProgram)
			,	m_tValue( T( 0))
		{
		}
		virtual ~GlFrameVariable()
		{
		}

	public:
		/**
		 * Assigns and activate the frame variable
		 */
		virtual bool Apply()
		{
			bool l_bReturn = true;

			if (m_bPresentInProgram)
			{
				if (m_glIndex == GL_INVALID_INDEX)
				{
					GetVariableLocation( m_strName.char_str());
				}

				if (m_glIndex != GL_INVALID_INDEX)
				{
					l_bReturn = VariableApplyer<T>::Apply( this);
				}
				else
				{
					m_bPresentInProgram = false;
				}
			}

			return l_bReturn;
		}
		/**
		* Defines the value of the variable
		*@param p_iValue : [in] The new value
		 */
		inline void SetValue( T p_tValue)
		{
			m_tValue = p_tValue;
			m_bChanged = true;
		}

		inline int		GetValue	()const	{ return m_tValue; }
	};
	template <> struct VariableApplyer<int>
	{
		static bool Apply( GlFrameVariable<int> * p_pVariable)
		{
			return OpenGl::Uniform1i( p_pVariable->GetGlIndex(), p_pVariable->GetValue());
		}
	};
	template <> struct VariableApplyer<float>
	{
		static bool Apply( GlFrameVariable<float> * p_pVariable)
		{
			return OpenGl::Uniform1f( p_pVariable->GetGlIndex(), p_pVariable->GetValue());
		}
	};

	class GlShaderObject
	{
	protected:
		friend class GlShaderProgram;
		static GLenum				ShaderTypes[eSHADER_OBJECT_TYPE_COUNT];
		GlShaderProgram			*	m_pParent;
		eSHADER_OBJECT_TYPE			m_eType;				//!< The program type
		wxString					m_strSource;			//!< ASCII Source-Code
		wxString					m_pathFile;				//!< File containing Source-Code
		bool						m_bCompiled;			//!< Tells if the program is compiled
		bool						m_bError;				//!< Tells if the program is in error after compilation
		GLuint						m_shaderObject;			//!< Shader Object
		wxString					m_compilerLog;
		bool						m_bAttached;

	protected:
		/**
		 * Constructor
		 */
		GlShaderObject( GlShaderProgram * p_pParent, eSHADER_OBJECT_TYPE p_eType);
		/**
		 * Destructor
		 */
		~GlShaderObject();

	public:
		/**
		 * Loads a shader file.
		 *@param p_filename : [in] The name of the ASCII file containing the shader.
		 *@return returns 0 if everything is ok.\n -1: File not found\n -2: Empty File\n -3: no memory
		 */
		void SetFile( const wxString & p_filename);
		/**
		 * Load program from text
		 *@param p_program : [in] Address of the memory containing the shader program.
		 */
		void SetSource( const wxString & p_strSource);
		/**
		 * Creates the program
		 */
		void CreateProgram();
		/**
		 * Destroys the program
		 */
		void DestroyProgram();
		/**
		 * Compiles program
		 */
		bool Compile();
		/**
		 * Attaches this shader to the given program
		 */
		void AttachTo( GlShaderProgram * p_pProgram);
		/**
		 * Detaches this shader from it's program
		 */
		void Detach();
		/**
		 * Get compiler messages
		 */
		void RetrieveCompilerLog( wxString & p_strCompilerLog);

		/**@name Accessors */
		//@{
		inline eSHADER_OBJECT_TYPE	GetShaderType			()const	{ return m_eType; }
		inline bool					IsCompiled				()const	{ return m_bCompiled; }
		inline wxString &			GetSource				()		{ return m_strSource; }
		inline const wxString &		GetSource				()const	{ return m_strSource; }
		inline wxString &			GetFile					()		{ return m_pathFile; }
		inline const wxString &		GetFile					()const	{ return m_pathFile; }
		inline const wxString &		GetCompilerLog			()const { return m_compilerLog; }
		//@}
	};

	class GlShaderProgram
	{
	protected:
		template <class Ty> friend struct FrameVariableCreator;

		bool											m_bLinked;					//!< Tells if the program is linked
		bool											m_bError;
		bool											m_bEnabled;					//!< Tells whether or not the shader is enabled
		GlShaderObject								*	m_pShaders[eSHADER_OBJECT_TYPE_COUNT];
		wxString										m_pathFile;
		FrameVariablePtrList							m_listFrameVariables;
		std::map<wxString, GlFrameVariable<int> *>		m_mapIntFrameVariables;
		std::map<wxString, GlFrameVariable<float> *>	m_mapFloatFrameVariables;
		GLuint											m_programObject;
		wxString										m_linkerLog;

	public:
		/**
		 * Constructor
		 */
		GlShaderProgram( const wxArrayString & p_shaderFiles);
		/**
		 * Constructor
		 */
		GlShaderProgram();
		/**
		 * Destructor
		 */
		~GlShaderProgram();
		/**
		 * Cleans the program up
		 */
		void Cleanup();
		/**
		 * Initialises the program
		 */
		bool Initialise();
		/**
		 * Creates a shader object
		 */
		void CreateObject( eSHADER_OBJECT_TYPE p_eType);
		/**
		 * Destroys a shader object
		 */
		void DestroyObject( eSHADER_OBJECT_TYPE p_eType);
		/**
		 * Begins the program
		 */
		bool Activate();
		/**
		 * Ends the program
		 */
		void Deactivate();
		/**
		 * Sets the file for the given program
		 */
		void SetProgramFile( const wxString & p_strFile, eSHADER_OBJECT_TYPE p_eType);
		/**
		 * Sets the text for the given program
		 */
		void SetProgramText( const wxString & p_strFile, eSHADER_OBJECT_TYPE p_eType);
		/**
		 * Retrieves the shader program wanted, creates it if null
		 */
		GlShaderObject * GetProgram( eSHADER_OBJECT_TYPE p_eType);
		/**
		 * Sets all objects file
		 */
		void SetFile( const wxString & p_path);
		/**
		 * Links all Shaders held by the program
		 */
		bool Link();
		/**
		 * Resets compilation variables to be able to compile again
		 */
		void ResetToCompile();
		/**
		 * Pass all the variables to the shader objects
		 */
		bool ApplyAllVariables();
		/**
		 * Retrieves Vertex Attribute from the program
		 */
		int GetAttributeLocation( const wxString & p_strName)const;
		/**
		 * Get Linker Messages
		 */
		void RetrieveLinkerLog( wxString & strLog);
		/**
		 * Creates a frame variable of wanted type
		 */
		GlFrameVariable<float> * CreateFloatFrameVariable( const wxString & p_strName, float p_fValue);
		GlFrameVariable<int> * CreateIntFrameVariable( const wxString & p_strName, int p_iValue);

		void SetFloatFrameVariableValue( const wxString & p_strName, float p_fValue);
		void SetIntFrameVariableValue( const wxString & p_strName, int p_iValue);

		/**@name Accessors */
		//@{
		inline	void	Enable			()	{ m_bEnabled = true; }
		inline	void	Disable			()	{ m_bEnabled = false; }

		inline	bool									IsLinked				()const { return m_bLinked; }
		inline	GLuint									GetProgramObject		()const { return m_programObject; }
		inline	const wxString &						GetLinkerLog			()const { return m_linkerLog; }
				FrameVariablePtrList::const_iterator	GetFrameVariablesBegin	()const { return m_listFrameVariables.begin(); }
				FrameVariablePtrList::iterator			GetFrameVariablesBegin	()		{ return m_listFrameVariables.begin(); }
				FrameVariablePtrList::const_iterator	GetFrameVariablesEnd	()const	{ return m_listFrameVariables.end(); }
				GlFrameVariable<int> *					GetIntFrameVariable		( const wxString & p_strName)const;
				GlFrameVariable<float> *				GetFloatFrameVariable	( const wxString & p_strName)const;
		//@}

	protected:
		GlShaderObject * DoCreateObject( eSHADER_OBJECT_TYPE p_eType);
	};
}

//*************************************************************************************************

#endif
