/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___Pixel___
#define ___Pixel___

#include "Point.h"

namespace ProceduralTextures
{
	template <typename T>
	class Pixel
	{
	public:
		T r;	//!< Red component
		T g;	//!< Green component
		T b;	//!< Blue component
		T a;	//!< Alpha component

	public:
		Pixel();
		template <typename U> Pixel( U p_r, U p_g, U p_b, U p_a);
		template <typename U> Pixel( const Pixel<U> & p_pxl);

		template <typename U> Pixel & operator =	( const Pixel<U>	&	p_px);
		template <typename U> Pixel & operator +=	( const Pixel<U>	&	p_px);
		template <typename U> Pixel & operator +=	( const U			&	p_t);
		template <typename U> Pixel & operator -=	( const Pixel<U>	&	p_px);
		template <typename U> Pixel & operator -=	( const U			&	p_t);
		template <typename U> Pixel & operator *=	( const Pixel<U>	&	p_px);
		template <typename U> Pixel & operator *=	( const U			&	p_t);
		template <typename U> Pixel & operator /=	( const Pixel<U>	&	p_px);
		template <typename U> Pixel & operator /=	( const U			&	p_t);
							  Pixel & operator >>	( size_t				p_t);

		template <typename U>					void		Set	( U p_r, U p_g, U p_b, U p_a);
		template <typename U>					void		Set	( const Pixel<U> & p_px);
		template <size_t Count1, size_t Count2>	void		Set	( const T * p_pBuffer1, const T * p_pBuffer2);
		template <typename U>					U			Sum	()const;
		template <typename U>					Pixel<U>	Mul	( const Pixel<U> & p_px)const;
		template <typename U>					Pixel<U>	Mul	( const U & p_val)const;
	};

	template <typename T, typename U> bool		operator ==	( const	Pixel<T> & p_pixel, const Pixel<U>	&	p_px);
	template <typename T, typename U> Pixel<T>	operator +	( const	Pixel<T> & p_pixel, const Pixel<U>	&	p_px);
	template <typename T, typename U> Pixel<T>	operator +	( const	Pixel<T> & p_pixel, const U			&	p_t);
	template <typename T, typename U> Pixel<T>	operator -	( const	Pixel<T> & p_pixel, const Pixel<U>	&	p_px);
	template <typename T, typename U> Pixel<T>	operator -	( const	Pixel<T> & p_pixel, const U			&	p_t);
	template <typename T, typename U> Pixel<T>	operator /	( const	Pixel<T> & p_pixel, const Pixel<U>	&	p_px);
	template <typename T, typename U> Pixel<T>	operator /	( const	Pixel<T> & p_pixel, const U			&	p_t);
	template <typename T, typename U> Pixel<T>	operator *	( const	Pixel<T> & p_pixel, const Pixel<U>	&	p_px);
	template <typename T, typename U> Pixel<T>	operator *	( const	Pixel<T> & p_pixel, const U			&	p_t);

	template <> class Policy<ProceduralTextures::UbPixel>
	{
	private:
		typedef ProceduralTextures::UbPixel value_type;

	public:
		static value_type	zero			()							{ return value_type(); }
		static value_type	unit			()							{ return value_type( 1, 1, 1, 1); }
		static void			init			( value_type & p_a)			{ p_a = zero(); }
		static bool			is_null			( const value_type & p_a)	{ return equals( p_a, zero()); }
		static value_type	negate			( const value_type & p_a)	{ return value_type( -p_a.r, -p_a.g, -p_a.b, -p_a.a); }
		static value_type	ass_negate		( value_type & p_a)			{ return assign( p_a, negate( p_a)); }
		static value_type	stick			( value_type & p_a)			{ return p_a; }

		template <typename Ty> static value_type	convert			( const	Ty			& p_value)					{ return value_type( p_value); }
		template <typename Ty> static bool			equals			( const	value_type	& p_a,	const Ty	& p_b)	{ return false; }
		template <typename Ty> static value_type	add				( const	value_type	& p_a,	const Ty	& p_b)	{ return p_a + convert<Ty>( p_b); }
		template <typename Ty> static value_type	substract		( const	value_type	& p_a,	const Ty	& p_b)	{ return p_a - convert<Ty>( p_b); }
		template <typename Ty> static value_type	multiply		( const	value_type	& p_a,	const Ty	& p_b)	{ return p_a * convert<Ty>( p_b); }
		template <typename Ty> static value_type	divide			( const	value_type	& p_a,	const Ty	& p_b)	{ return p_a / convert<Ty>( p_b); }
		template <typename Ty> static value_type	ass_add			(		value_type	& p_a,	const Ty	& p_b)	{ return assign( p_a, add<Ty>( p_a, p_b)); }
		template <typename Ty> static value_type	ass_substract	(		value_type	& p_a,	const Ty	& p_b)	{ return assign( p_a, substract<Ty>( p_a, p_b)); }
		template <typename Ty> static value_type	ass_multiply	(		value_type	& p_a,	const Ty	& p_b)	{ return assign( p_a, multiply<Ty>( p_a, p_b)); }
		template <typename Ty> static value_type	ass_divide		(		value_type	& p_a,	const Ty	& p_b)	{ return assign( p_a, divide<Ty>( p_a, p_b)); }
		template <typename Ty> static value_type	assign			(		value_type	& p_a,	const Ty	& p_b)	{ return p_a = convert<Ty>( p_b); }
	};

	template <typename T>
	class Buffer
	{
	private:
		T * m_pPixels;							//!< The contiguous buffer of values
		bool m_bOwnsPixels;						//!< Tells if the buffer owns it's buffer
		Point<size_t, 2> m_ptSize;				//!< The dimensions of the buffer
		std::vector< DynPoint<T> > m_arrayRows;	//!< The array of rows

	public:
		Buffer( const Point<size_t, 2> & p_ptSize);
		Buffer( const Buffer & p_pixelBuffer);
		virtual ~Buffer();

		Buffer & operator =( const Buffer & p_pixelBuffer);
		Buffer & operator =( const T * p_pBuffer);

		void	clear();
		void	init();
		void	init( const Point<size_t, 2> & p_ptSize);
		void	swap( Buffer & p_pixelBuffer);
		void	link( T * p_pBuffer);
		void	unlink();
		template <typename U, size_t Count1, size_t Count2, bool InvertedX, bool InvertedY >
		void	set( const U * p_pBuffer1, const U * p_pBuffer2);

		inline const DynPoint<T> &	operator []	( size_t p_uiIndex)const	{ return m_arrayRows[p_uiIndex]; }
		inline DynPoint<T> &		operator []	( size_t p_uiIndex)			{ return m_arrayRows[p_uiIndex]; }
		inline const T *			const_ptr	()const						{ return & m_pPixels[0]; }
		inline T *					ptr			()							{ return & m_pPixels[0]; }
		inline size_t				size		()const						{ return m_ptSize[0] * m_ptSize[1] * sizeof( T); }
		inline size_t				count		()const						{ return m_ptSize[0] * m_ptSize[1]; }
	};
}

#include "Pixel.inl"

#endif
